@@ -5,14 +5,15 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-from tensorflow.python.keras import backend as K
+from keras import backend as K
 graph_unique_name = K.get_graph().unique_name
 
-from tensorflow.python.keras.layers import Dense
-from tensorflow.python.keras.layers import Activation
-from tensorflow.python.keras.layers import Concatenate
-from tensorflow.python.keras.layers import Lambda
-from tensorflow.python.keras.models import Model
+from keras.layers import Dense
+from keras.layers import Activation
+from keras.layers import Concatenate
+from keras.layers import Lambda
+# from keras.layers import BatchNormalization
+from keras.models import Model
 from tensorflow import tensordot, expand_dims
 
 from ..utils import to_list, unpack_singleton, is_same_tensor, unique_tensors
@@ -79,11 +80,21 @@ def Functional(
         hidden_layers = []
     else:
         hidden_layers = to_list(hidden_layers)
+    if not all([isinstance(n, int) for n in hidden_layers]):
+        raise TypeError("Enter a list of integers as the third input assigning layer widths, e.g. [10,10,10]. ")
     # prepare kernel initializers.
-    activations, def_biasinit, def_kerinit = \
-        prepare_default_activations_and_initializers(
-        len(hidden_layers) * [activation] + [output_activation]
-    )
+    if isinstance(activation, list) and isinstance(activation[0], list):
+        assert len(activation) == len(hidden_layers), \
+            'to specify activation per layer, make sure len(activation)==len(hidden_layer).'
+        activations, def_biasinit, def_kerinit = \
+            prepare_default_activations_and_initializers(
+            activation + [output_activation]
+        )
+    else:
+        activations, def_biasinit, def_kerinit = \
+            prepare_default_activations_and_initializers(
+            len(hidden_layers) * [activation] + [output_activation]
+        )
     if kernel_initializer is None:
         kernel_initializer = def_kerinit
     elif isinstance(kernel_initializer, (float, int)):
@@ -128,7 +139,7 @@ def Functional(
                 )
             )
         elif validations.is_field(fld):
-            output_fields.append(fields)
+            output_fields.append(fld)
         else:
             raise TypeError(
                 'Please provide a "list" of field names of'
@@ -179,7 +190,7 @@ def Functional(
             )
             res_output = layers[-1](net_input)
             # Apply the activation.
-            if activations[0].activation.__name__ != 'linear':
+            if activations[0].activation_name != 'linear':
                 layers.append(activations[0])
                 res_outputs.append(layers[-1](res_output))
         net[-1] = res_outputs[-1]
@@ -198,8 +209,12 @@ def Functional(
         )
         layers.append(layer)
         net[-1] = layer(net[-1])
+        # # Add batch normalization. 
+        # layer = BatchNormalization()
+        # layers.append(layer)
+        # net[-1] = layer(net[-1])
         # Apply the activation.
-        if activations[nLay].activation.__name__ != 'linear': #nLay<len(hidden_layers)-1 and
+        if activations[nLay].activation_name != 'linear':
             layer = activations[nLay]
             layers.append(layer)
             net[-1] = layer(net[-1])
@@ -222,7 +237,7 @@ def Functional(
         last_layers = [out]
         last_output = out(net_output)
         # add the activation on the output.
-        if activations[-1].activation.__name__ != 'linear':
+        if activations[-1].activation_name != 'linear':
             layer = activations[-1]
             last_layers.append(layer)
             last_output = layer(last_output)
