@@ -602,16 +602,37 @@ def atan(x):
     return _apply_function(x, 'atan')
 
 
-def atan2(x):
-    """Computes atan2 of x element-wise.
+def atan2(y, x):
+    """Computes atan2 of y, x pair, element-wise.
 
     # Arguments
+        y: Functional object.
         x: Functional object.
 
     # Returns
         A new functional object.
     """
-    return _apply_function(x, 'atan2')
+    validate_functional(x)
+    validate_functional(y)
+
+    fun = get_activation('atan2')
+    lmbd, outputs = [], []
+    for i in range(len(x.outputs)):
+        lmbd.append(
+            Lambda(
+                lambda xs: fun(xs[0], xs[1]),
+                name=graph_unique_name("{}".format('atan2'))
+            )
+        )
+        outputs += [lmbd[-1]([yi, xi]) for yi, xi in zip(y.outputs, x.outputs)]
+    Functional = x.get_class()
+    inputs = y.inputs.copy() + x.inputs.copy()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = outputs,
+        layers = lmbd
+    )
+    return res
 
 
 def cot(x):
