@@ -11,23 +11,23 @@ from numpy import pi
 
 import tensorflow as tf
 from tensorflow.python import keras as k
-from tensorflow.python.keras import backend as K
+from keras import backend as K
 
 # interface for some keras features to be acessible across sciann.
-from tensorflow.python.keras.backend import is_keras_tensor as is_tensor
-from tensorflow.python.keras.backend import floatx
-from tensorflow.python.keras.backend import set_floatx
-from tensorflow.python.keras.utils.vis_utils import model_to_dot
-from tensorflow.python.keras.utils.vis_utils import plot_model
-from tensorflow.python.keras.initializers import RandomUniform as default_bias_initializer
-from tensorflow.python.keras.initializers import GlorotNormal as default_kernel_initializer
-from tensorflow.python.keras.initializers import Constant as default_constant_initializer
-from tensorflow.python.keras.initializers import VarianceScaling
-from tensorflow.python.keras.regularizers import l1_l2
+from keras.backend import is_keras_tensor as is_tensor
+from keras.backend import floatx
+from keras.backend import set_floatx
+from keras.utils.vis_utils import model_to_dot
+from keras.utils.vis_utils import plot_model
+from keras.initializers import RandomUniform as default_bias_initializer
+from keras.initializers import GlorotNormal as default_kernel_initializer
+from keras.initializers import Constant as default_constant_initializer
+from keras.initializers import VarianceScaling
+from keras.regularizers import l1_l2
 
 from .initializers import SciKernelInitializer as KInitializer
 from .initializers import SciBiasInitializer as BInitializer
-from .activations import get_activation, SciActivation, SciActivationLayer
+from .activations import get_activation, SciActivation, SciActivationLayer, SciRowdyActivationLayer
 from .validations import is_functional
 
 from pybtex.database.input import bibtex
@@ -77,10 +77,10 @@ def get_default_log_path():
     return _DEFAULT_LOG_PATH
 
 
-def initialize_bib(bib_file):
+def initialize_bib(bib_str):
     global _BIBLIOGRAPHY
     global _BIBLIOGRAPHY_TO_OUTPUT
-    _BIBLIOGRAPHY = bibtex.Parser().parse_file(bib_file)
+    _BIBLIOGRAPHY = bibtex.Parser().parse_string(bib_str)
     _BIBLIOGRAPHY_TO_OUTPUT.append(_BIBLIOGRAPHY.entries['haghighat2021sciann'])
     _BIBLIOGRAPHY_TO_OUTPUT.append(_BIBLIOGRAPHY.entries['raissi2019physics'])
 
@@ -205,24 +205,32 @@ def prepare_default_activations_and_initializers(actfs, seed=None):
     bias_initializer = []
     kernel_initializer = []
     for lay, actf in enumerate(to_list(actfs)):
-        if isinstance(actf, str):
-            lay_actf = actf.lower().split('-')
-            f = get_activation(lay_actf[-1])
-        elif callable(actf):
-            lay_actf = actf.__name__.lower().split('-')
-            f = actf
-        else:
-            raise ValueError('expected a string for actf: {}'.format(actf))
-
-        bias_initializer.append(BInitializer(lay_actf[-1], lay, seed))
-        kernel_initializer.append(KInitializer(lay_actf[-1], lay, seed))
+        # initializers.
+        bias_initializer.append(BInitializer(lay, seed))
+        kernel_initializer.append(KInitializer(lay, seed))
         w = kernel_initializer[-1].w0
-
-        if len(lay_actf) == 2:
-            append_to_bib("jagtap2020locally")
-            activations.append(SciActivationLayer(w, f, lay_actf[0]))
+        # support rowdy net.
+        layer_actfs = []
+        for sig in to_list(actf):
+            if isinstance(sig, str):
+                lay_actf = sig.lower().split('-')
+                f = get_activation(lay_actf[-1])
+            elif callable(sig):
+                lay_actf = sig.__name__.lower().split('-')
+                f = sig
+            else:
+                raise TypeError('expected a string for actf: {}'.format(sig))
+            if len(lay_actf) == 2:
+                append_to_bib("jagtap2020locally")
+                layer_actfs.append(SciActivationLayer(w, f, lay_actf[0]))
+            else:
+                layer_actfs.append(SciActivation(w, f))
+        if len(layer_actfs) == 1:
+            sig_layer = layer_actfs[0]
         else:
-            activations.append(SciActivation(w, f))
+            append_to_bib("jagtap2021deep")
+            sig_layer = SciRowdyActivationLayer(layer_actfs)
+        activations.append(sig_layer)
 
     return activations, bias_initializer, kernel_initializer
 
